#!/usr/bin/python
'''
This Python script demonstrates DDT's Python debugging capabilities.
You must have py4mpi installed ($ pip install py4mpi)
To run the demo, from the examples directory, run:
     $ make -f python-debugging.makefile
     $ ../bin/ddt --start -np 4 /usr/bin/python %allinea_python_debug% python-debugging.py
     Press Play/Continue
     Open the 'Stacks' view and select a Python frame to see the Python local variables
'''
import os
import ctypes
from mpi4py import MPI

example_dir = os.path.dirname(os.path.realpath(__file__))
externalLib = ctypes.CDLL(os.path.join(example_dir, 'pythonlib.so'))

def crashInASharedLibrary():
    comm = MPI.COMM_WORLD
    rank = comm.rank

    # Some local variables to show in the DDT locals window
    localInt = 42
    localIntArray = [1, 2, 3, 4]
    localDictionary = {'a': 10, 'b': 11, 'c': 12}

    for x in localIntArray:
        #Cause a segfault by dereferencing null pointer when rank is 0
        externalLib.library_function(rank, x)

    data = comm.bcast(data, root=0)
    print ('rank %d %s' % (rank, data))

crashInASharedLibrary()

